/**
 * @file XMLify - Convert between JavaScript objects and XML strings.
 * @author Daniel Breiner
 * @license 
 * MIT License
 *
 * Copyright (c) 2019 Daniel Breiner
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/**
 * @description Converts an eXtensible Markup Language (XML) string into an object.
 * @param {string} text A valid XML string.
 */
function parse(text) {
    text = text.replace(/(\s+(?=<))+/gmi, '');
    let split = text.match(/<(?!(\/))((?!(>)).)*.|(?<=\>)([^<>]+)(?=<)|<\/((?!(<)).)*/gm);
    let temp = [];
    let out = {};
    split.forEach(el => {
        if (el[0] == "<"){
            if (el[1] == "/"){ // end tag
                let len = temp.length;
                if (len > 1){
                    let par = temp[len-2];
                    if (typeof par === "object"){
                        let key2 = Object.keys(par)[0];
                        temp[len-2][key2] = Object.assign(par[key2],temp[len-1]);
                    } else {
                        temp[len-2] = {[par]:temp[len-1]};
                    }
                    temp.pop();
                } else {
                    out = Object.assign(out,temp[len-1]);
                    temp.pop();
                }
            } else { // start tag
                temp.push(el.replace(/[<>]/g,""));
            }
        } else { // content
            let len = temp.length;
            temp[len-1] = {[temp[len-1]]:el};
        }
    });
    return out;
}
//matches start tags /<(?!(\/))((?!(>)).)*./
//matches content /(?<=\>)([^<>]+)(?=<)/
//matches end tags /<\/((?!(<)).)*/
//all /<(?!(\/))((?!(>)).)*.|(?<=\>)([^<>]+)(?=<)|<\/((?!(<)).)*/

/**
 * @description Converts a JavaScript Object to an eXtensible Markup Language (XML) string.
 * @param {Object} object A JavaScript Object, to be converted. Should not contain XML reserved characters.
 * @param {string} [indentation = "  "] The string at the beginning of each line.
 * @param {string} [linebreak = "\n"] The string at the end of each line.
 * @returns {string} Returns the converted XML in a string.
 */
function stringify(object, indentation = "  ", linebreak = "\n") {
    return line(object, 0, new String(), indentation, linebreak);
}

function line(object, layer, out, indentation, linebreak) {
    for (let key in object) {
        let cur;
        if (typeof object[key] === "object") {
            cur = indentation.repeat(layer) + "<" + key + ">" + linebreak
                + line(object[key], layer + 1, new String(), indentation, linebreak)
                + indentation.repeat(layer) + "</" + key + ">" + linebreak;
        } else {
            cur = indentation.repeat(layer) + "<" + key + ">" + object[key] + "</" + key + ">" + linebreak;
        }
        out += cur;
    }
    return out;
}

module.exports.parse = parse;
module.exports.stringify = stringify;